# -*- coding: utf-8 -*-

"""Created on 16.11.2019

.. moduleauthor:: Paweł Pecio
"""
import logging
import warnings
from datetime import datetime
from importlib import import_module
from types import MappingProxyType
from typing import Tuple, Mapping, MutableMapping, Union

from patchwork.client.task import Task


class TaskDecodeError(ValueError):
    """
    Received message cannot be decoded into task.
    """

    def __init__(self, payload: bytes, meta: Mapping, exc: Exception, handled: bool):
        """
        :param payload: Received payload
        :param meta:    Received meta
        :param exc:     Raised exception
        :param handled: True if exception has been handled by client and this is just a notification exception,
                        handled means that task is not lost on the queue. If False, this case is not handled
                        and worker should take care of logging and administrator notification in the same
                        way as other task processing messages are handled.
        """
        self.payload = payload
        self.meta = meta
        self.exception = exc
        self.is_handled = handled

    def __str__(self):
        return ("Handled" if self.is_handled else "Unhandled") + f" {self.__class__.__name__}({self.exception})"


class ClientCommon:
    """
    Common code for sync and async clients
    """

    asynchronous: bool

    def __init__(self, *, name: str = None, serializer: Union[str, type, object] = 'json'):
        """
        :param name: Client name
        :param serializer:  task serializer class, instance or module, must implement dumps and loads methods.
                            Can be a path, like: package.module[:class].
                            Default is standard json module.
        """
        self._name = name if name is not None else id(self)
        self._serializer = serializer
        self.logger = logging.getLogger(f'patchwork.client.{self.name}')
        self._is_started = False

    @property
    def name(self):
        return self._name

    @property
    def serializer(self):
        if isinstance(self._serializer, str):
            mod_name, class_name = self._serializer.split(':')
            mod = import_module(mod_name)
            if class_name:
                mod = getattr(mod, class_name)

            assert hasattr(mod, 'dumps') and hasattr(mod, 'loads'), \
                f"Invalid serializer '{self._serializer}'. No dumps() or loads() method found in {mod}"

            self._serializer = mod

        return self._serializer

    def _serialize_task(self, task: Task) -> Tuple[bytes, Mapping]:
        """
        Take task and returns serialized task payload as bytes and task metadata as mapping.
        :param task:
        :return:
        """
        task.meta.scheduled = datetime.utcnow().timestamp()
        state = dict(task.__getstate__())
        meta = MappingProxyType(state.pop('meta'))
        payload: bytes = self.serializer.dumps(state)
        return payload, meta

    def _deserialize_task(self, payload: bytes, meta: Mapping) -> Task:
        try:
            state = self.serializer.loads(payload)
            state['meta'] = dict(meta)
            if 'received' not in meta:
                state['meta']['received'] = datetime.utcnow().timestamp()

            task = Task.__new__(Task)
            task.__setstate__(state)
            return task
        except Exception as exc:
            handled = self._handle_deserialize_error(payload, meta, exc)
            raise TaskDecodeError(payload, meta, exc, handled) from exc

    def _handle_deserialize_error(self, payload: bytes, meta: Mapping, exc: Exception) -> bool:
        """
        Called when exception is raised during message deserialization and task initialization.
        This method is responsible of making sure that task won't be lost, however simple resending
        to the queue might be not accurate because if error happens once it probably will happen
        every time.
        :param payload:
        :param meta:
        :param exc:
        :return: True if message has been somehow handled and worker should not take any action. Otherwise
                 return False, so worker may report this issue using own logging mechanism
        """
        return False

    def __repr__(self):
        res = super().__repr__()
        return f'<{self.name}: {res[1:-1]} [serializer={self.serializer.__name__}]>'

    def __del__(self):
        if self._is_started:
            warnings.warn(f"Destroying unstopped client may leads to data loss! stop() didn't finished or called")


class SynchronousClient(ClientCommon):
    """
    Synchronous patchwork base client. This is a fallback client for external synchronous code
    without async support.
    """

    is_asynchronous = False

    def start(self):
        self._is_started = True
        self._setup()

    def stop(self):
        self._teardown()
        self._is_started = False

    def _setup(self):
        pass

    def _teardown(self):
        pass

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.get()
        except EOFError:
            raise StopIteration()

    def send(self, task: Task, timeout: float = None):
        """
        Sends given task, if operation fail in given timeout raises TimeoutError.
        :param task:
        :param timeout: send operation timeout, None means no timeout
        :return:
        """
        payload, meta = self._serialize_task(task)
        self._send(payload, meta, timeout)

    def get(self, timeout: int = None):
        """
        Waits given time for next task and when arrived returns it. If no task came raises EOFError.
        :param timeout:
        :return:
        """
        try:
            payload, meta = self._fetch_one(timeout)
        except TimeoutError:
            raise EOFError()

        return self._deserialize_task(payload, meta)

    def commit(self, task, timeout: int = None):
        """
        Commits given task returned by get() method. If operation fail in given timeout
        raises TimeoutError.
        :param task:
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    def rollback(self, task, timeout: int = None):
        """
        Rollback given task returned by get() method. If operation fail in given timeout
        raises TimeoutError.
        :param task:
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    def _send(self, payload: bytes, meta: Mapping, timeout: float = None):
        """
        Sends given payload
        :param payload: Task payload to send
        :param meta:    Task metadata, note: task metadata should be send along with the payload message,
                        if transport does not support metadata serialize it and send with payload (eg. as
                        serialized metadata length + metadata bytes + payload bytes)
        :param timeout: Requested send operation timeout, None means no timeout
        :return:
        """
        raise NotImplementedError()

    def _fetch_one(self, timeout) -> Tuple[bytes, Mapping]:
        """
        Fetch one incoming message from the queue.
        :param timeout: Waits given time for message, if there is no message in given timeout raises TimeoutError.
                        None means no timeout, `0` means no wait.
        :return:
        """
        raise NotImplementedError()


class AsyncClient(ClientCommon):
    """
    Patchwork asynchronous client
    """

    is_asynchronous = True

    async def start(self):
        """
        Initialize client here, this method should be called just before first get() or send() call.
        Start method should not fail for already started client or if client has been stopped.
        Client can be started and stopped multiple times.
        :return:
        """
        self._is_started = True
        await self._setup()

    async def stop(self):
        """
        Finish and free client resources, this method should be called just after last get() or send() call.
        Stop method should not fail for already stopped client.
        Client can be started and stopped multiple times.
        :return:
        """
        await self._teardown()
        self._is_started = False

    async def _setup(self):
        pass

    async def _teardown(self):
        pass

    async def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self.get()
        except EOFError:
            raise StopAsyncIteration()

    async def send(self, task, timeout: float = None):
        """
        Sends given task, if operation fail in given timeout raises TimeoutError.
        When this method returns it's considered as task has been delivered successfully and it's guaranteed
        that won't be lost (eg queue backend committed the message)
        :param task:
        :param timeout: send operation timeout, None means no timeout. 0 means immediatelly
        :return:
        """
        payload, meta = self._serialize_task(task)
        await self._send(payload, meta, timeout)

    async def get(self, timeout: float = None):
        """
        Waits given time for next task and when arrived returns it. If no task came raises EOFError.
        :param timeout:
        :return:
        """
        try:
            payload, meta = await self._fetch_one(timeout)
        except TimeoutError:
            raise EOFError()

        task = self._deserialize_task(payload, meta)
        task.receiver = self
        return task

    async def commit(self, task, timeout: float = None):
        """
        Commits given task returned by get() method. If operation fail in given timeout.
        Meaning of "commit" depends on queue backend, but should be considered as success of task handling
        (which is not the same as success of task execution). Success of task handling means that task
        has been executed successfully or rescheduled successfully and given task can be safely removed
        from the queue. Whatever happens task won't be lost.
        raises TimeoutError.
        :param task:
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    async def _send(self, payload: bytes, meta: Mapping, timeout: float = None):
        """
        Sends given payload
        :param payload: Task payload to send
        :param meta:    Task metadata, note: task metadata should be send along with the payload message,
                        if transport does not support metadata serialize it and send with payload (eg. as
                        serialized metadata length + metadata bytes + payload bytes)
        :param timeout: Requested send operation timeout, None means no timeout, 0 means immediatelly
        :return:
        """
        raise NotImplementedError()

    async def _fetch_one(self, timeout: float = None) -> Tuple[bytes, MutableMapping]:
        """
        Fetch one incoming message from the queue.
        :param timeout: Waits given time for message, if there is no message in given timeout raises TimeoutError.
                        None means no timeout, `0` means no wait.
        :return:
        """
        raise NotImplementedError()

