# -*- coding: utf-8 -*-

"""Created on 16.08.2019

.. moduleauthor:: Paweł Pecio
"""
import uuid
import weakref
from dataclasses import dataclass, asdict, fields
from datetime import datetime, timezone
from types import MappingProxyType
from typing import Mapping, Union, Tuple, Optional


@dataclass
class TaskMetadata:
    # task should not be executed before given Unix timestamp (UTC), in seconds
    not_before: Optional[float] = None

    # task expires after given Unix timestamp (UTC), in seconds; worker should skip task
    expires: Optional[float] = None

    # max number of retries, if not given use worker defaults
    max_retry: Optional[int] = None

    # current number of task run attempt, incremented by one each time task is handled
    attempt: int = 0

    # on sender side tells client which queue is preferred, on worker side it's a name of queue task came from
    queue_name: Optional[bytes] = None

    # field filled by client send method, task scheduled timestamp (UTC), in seconds
    scheduled: float = None

    # field filled by worker receiver, tells when task has been received
    received: float = None

    # additional information passed by receiver
    info: Optional[Mapping] = None


class Task:
    """
    Task message
    """

    __slots__ = ('_id', '_meta', '_msg_type', '_args', '_kwargs', '_receiver')

    _protocol_version = (1, 0)

    def __init__(self, msg_type: bytes, args: Tuple = None, kwargs: Mapping = None, *,
                 not_before: Union[datetime, float] = None, expires: Union[datetime, float] = None,
                 max_retry: int = None, queue_name: bytes = None):

        if not_before is not None:
            if isinstance(not_before, datetime):
                not_before = not_before.replace(tzinfo=timezone.utc).timestamp()

            if not_before < 0:
                raise ValueError("not_before cannot be negative")

        if expires is not None:
            if isinstance(expires, datetime):
                expires = expires.replace(tzinfo=timezone.utc).timestamp()

            if expires < 0:
                raise ValueError("expires cannot be negative")

        if max_retry is not None and max_retry < 0:
            raise ValueError("max_retry cannot be negative")

        self._id = str(uuid.uuid4())
        self._meta = TaskMetadata(not_before=not_before, expires=expires, max_retry=max_retry,
                                  queue_name=queue_name)
        self._msg_type = msg_type
        self._args = args if args is not None else tuple()
        self._kwargs = kwargs if kwargs is not None else MappingProxyType({})
        self._receiver = None

    def __repr__(self):
        return f"<Task {self._id} [{self._msg_type.decode('latin-1')}]>"

    def __getstate__(self) -> Mapping:
        """
        Returns task state in python simple data types which can be easily serialized using
        standard libraries
        :return:
        """

        return {
            'id': self._id,
            'meta': asdict(self._meta),
            'type': self._msg_type,
            'args': self._args,
            'kwargs': self._kwargs,
            'protocol': self._protocol_version
        }

    def __setstate__(self, state: Mapping):
        """
        Restore task from given state
        :param state:
        :return:
        """
        if not self.is_compatible(state['protocol']):
            raise TypeError(f"Given protocol version {state['protocol']} is not supported")

        meta_fields = fields(TaskMetadata)

        # extract metadata dataclass fields from incoming task meta, all other fields put as
        # additional information using metadata info field
        meta = {}
        info = state['meta']

        for field in meta_fields:
            if field.name in info:
                meta[field.name] = info.pop(field.name)
        meta['info'] = info

        self._id = state['id']
        self._meta = TaskMetadata(**meta)
        self._msg_type = state['type']
        self._args = tuple(state['args'])
        self._kwargs = MappingProxyType(state['kwargs'])

    @property
    def receiver(self):
        """
        Returns client instance which receives this task. Available only
        for received tasks, otherwise None. Reference to client instance is weak
        so this property may return None also if task has been received but
        client instance is no longer available.
        :return:
        """
        return self._receiver() if self._receiver is not None else None

    @receiver.setter
    def receiver(self, c):
        self._receiver = weakref.ref(c)

    def is_compatible(self, protocol):
        """
        Returns if given protocol version is compatible and supported by this class
        :param protocol:
        :return:
        """
        return protocol <= self._protocol_version

    @property
    def meta(self) -> TaskMetadata:
        return self._meta

    @property
    def type(self) -> bytes:
        return self._msg_type

    def clone(self):
        # TODO:
        raise NotImplementedError()